/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 932 $ $LastChangedDate: 2002-09-11 17:06:10 +0200 (Wed, 11 Sep 2002) $
 *
 * Standard library for Gont, Stdio module C implementation.
 */

#include <gont/runtime.h>
#include <gont/array.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

void gont_fputs(FILE *f, struct gont_string *s, struct gont_charset *ch);

struct gont_file {
	FILE *fp;
	struct gont_charset *charset;
	int p_kind;
};

static struct gont_file *gont_make_file(FILE *f)
{
	struct gont_file *gf;

	gf = gont_xmalloc0(sizeof(*gf));
	gf->fp = f;
	gf->charset = gont_default_charset;

	return gf;
}

void Stdio__set_charset(struct gont_file *f, struct gont_charset *c)
{
	f->charset = c;
}

struct gont_file *Stdio___stdX(gont_int x)
{
	switch (x) {
	case 0:
		return gont_make_file(stdin);	
	case 1:
		return gont_make_file(stdout);	
	case 2:
		return gont_make_file(stderr);	
	default:
		gont_invalid_argument("Stdio___stdX");
		return NULL;	/* Not reached. */
	}
}

struct gont_file *Stdio__fopen(struct gont_string *filename, 
			       struct gont_string *flags)
{
	GONT_STRING_TO_CSTRING_N(filename, gont_default_charset);
	GONT_STRING_TO_CSTRING_N(flags, gont_default_charset);
	FILE *f = fopen(c_filename, c_flags);
	
	if (f == NULL)
		gont_sys_error(errno);
		
	return gont_make_file(f);
}

struct gont_file *Stdio__fdopen(gont_int fd, struct gont_string *flags)
{
	GONT_STRING_TO_CSTRING_N(flags, gont_default_charset);
	FILE *f = fdopen(fd, c_flags);
	
	if (f == NULL)
		gont_sys_error(errno);
		
	return gont_make_file(f);
}

struct gont_file *Stdio__popen(struct gont_string *cmd, 
			       struct gont_string *flags)
{
	GONT_STRING_TO_CSTRING_N(cmd, gont_default_charset);
	GONT_STRING_TO_CSTRING_N(flags, gont_default_charset);
	FILE *f = popen(c_cmd, c_flags);
	struct gont_file *r;
	
	if (f == NULL)
		gont_sys_error(errno);
		
	r = gont_make_file(f);
	r->p_kind = 1;

	return r;
}

gont_int Stdio__pclose(struct gont_file *f)
{
	int r;
	
	if (f->fp == NULL || f->p_kind == 0)
		gont_invalid_argument("Stdio::pclose");
	r = pclose(f->fp);
	f->fp = NULL;
	return r;
}

void Stdio__fclose(struct gont_file *f)
{
	int err;

	if (f->fp == NULL)
		gont_invalid_argument("Stdio::fclose");
	
	err = fclose(f->fp);
	if (err)
		err = errno;
	
	f->fp = NULL;
	
	if (err)
		gont_sys_error(err);
}

struct gont_string *Core__read_line_f(FILE *f);

struct gont_string *Stdio__fgets(struct gont_file *f)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::fgets");

	return Core__read_line_f(f->fp);
}

void Stdio__fputs(struct gont_file *f, struct gont_string *s)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::fputs");
	
	gont_fputs(f->fp, s, f->charset);
}

gont_int Stdio__feof(struct gont_file *f)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::feof");
	return feof(f->fp) ? 1 : 0;
}

gont_int Stdio__ferror(struct gont_file *f)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::ferror");
	return ferror(f->fp) ? 1 : 0;
}

void Stdio__clearerr(struct gont_file *f)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::clearerr");
	clearerr(f->fp);
}

gont_int Stdio__fileno(struct gont_file *f)
{
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::fileno");
	return fileno(f->fp);
}

/* TODO: fix it to read *characters* not bytes.  */
struct gont_string *Stdio__fread(struct gont_file *f, gont_int cnt)
{
	char buf[1024];
	char *ptr;
	
	if (f->fp == NULL)
		gont_invalid_argument("Stdio::fread");
	
	if (cnt < 1024)
		ptr = buf;
	else
		ptr = gont_xmalloc_atomic(cnt);
		
	if (fread(ptr, cnt, 1, f->fp) != 1)
		gont_sys_error(errno);
	
	return gont_string_of_cstring(ptr, cnt, f->charset);
}

/* TODO:
void fflush(struct gont_file *f);
void fseek(struct gont_file *f, int offset, fwhence whence);
int ftell(struct gont_file *f);
void rewind(struct gont_file *f);
void setlinebuf(struct gont_file *f);
void setnobuf(struct gont_file *f);
*/

/* File functions. */
void Stdio__remove(struct gont_string *filename)
{
	GONT_STRING_TO_CSTRING_N(filename, gont_default_charset);
	if (remove(c_filename))
		gont_sys_error(errno);
}

void Stdio__rename(struct gont_string *oldname, struct gont_string *newname)
{
	GONT_STRING_TO_CSTRING_N(oldname, gont_default_charset);
	GONT_STRING_TO_CSTRING_N(newname, gont_default_charset);
	if (rename(c_oldname, c_newname))
		gont_sys_error(errno);
}
