/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1012 $ $LastChangedDate: 2002-09-21 20:16:14 +0200 (Sat, 21 Sep 2002) $
 *
 * Xmldump - Mappings from some Ast types to Xml::node.
 *
 * This is currently used by Module. There is also Gont::bind_xml_types()
 * entry point, output of which can be used by iogen. In future we might
 * use it for debugging or something else.
 */
 
open Ast;
open Util;

Xml::node xmlize_tyvars(<id>list ids)
{
    def one(i)
      (Xml::Element("tyvar", Nil, [("name", i)]));
    
    return Xml::Element("tyvars", List::map(one, ids), Nil);
}

Xml::node xmlize_field(<<'a>symbol>field f)
{
    return Xml::Element("field", [xmlize_ty(f.fld_type)],
			[("name", simple_name(f.fld_symbol))]);
}

Xml::node xmlize_ty_int(ty t)
{
    <Xml::node>list children = Nil;
    <*(string, string)>list attrs = Nil;
    string element;

    def xmlize_parm(p)
    {
	switch p {
	case (tt, Some(n)):
	    return Xml::Element("parm", [xmlize_ty_int(tt)], [("name", n)]);
	case (tt, None):
	    return Xml::Element("parm", [xmlize_ty_int(tt)], []);
	}
    }
   
    // change unxmlize_ty() too!
    switch t {
    case Int:
	element = "int";
    
    case Void:
	element = "void";
	
    case Bool:
	element = "bool";
	
    case Char:
	element = "char";
	
    case Float:
	element = "float";
	
    case String:
	element = "string";
	
    case Abstract_type:
	element = "abstract-type";
    
    case Exn:
	element = "exn";
    
    case Tuple(lst):
	element = "tuple";
	attrs = [("length", itoa(List::length(lst)))];
	children = List::map(xmlize_ty_int, lst);

    case Func(ft):
	element = "func";
	let parms = Xml::Element("parms", List::map(xmlize_parm, ft.parms), []),
	    retty = Xml::Element("ret-type", [xmlize_ty_int(ft.ret_type)], []) in 
	children = [xmlize_tyvars(ft.tyvars), retty, parms];
    
    case Tyvar(x):
	element = "tyvar";
	attrs = [("name", x)];
    
    case Struct(lst):
	element = "struct";
	children = List::map(xmlize_field, lst);
	
    case Opt_struct(lst):
	element = "opt-struct";
	children = List::map(xmlize_field, lst);
	
    case Union(lst):
	element = "union";
	children = List::map(xmlize_field, lst);
    
    case Named_type(nt):
	let tys = List::map(xmlize_ty_int, nt.parms) in
	children = [Xml::Element("parms", tys, [])];
	attrs = [("name", qualified_name(nt.tydef))];
	element = "named-type";
    
    case X_tyvar(x):
	// this might be used for debug, it should not appear in
	// compiled interfaces and the like
	attrs = [("name", x.name)];
	element = "x-tyvar";
	switch x.t {
	case Some(tt):
	    children = [xmlize_ty_int(tt)];
	case None:
	    skip;
	}
    }

    return Xml::Element(element, children, attrs);
}

Xml::node xmlize_ty(ty tt)
{
    ty strip_current_module(ty x)
    {
	switch x {
	case Named_type(nt):
	    nt_record new_nt = {parms = nt.parms, tydef = nt.tydef};
	    
	    switch nt.tydef {
	    case Qualified_name(m, n):
		if (m == Util::current_module)
		    new_nt.tydef = Simple_name(n);
	    case Symbol(s):
		if (s.module == Util::current_module)
		    new_nt.tydef = Simple_name(s.name);
	    case _:
		skip;
	    }

	    return Named_type(new_nt);
	    
	case X_tyvar(xt):
	    switch xt.t {
	    case Some(t):
		return strip_current_module(t);
	    case None:
		return x;
	    }
	    
	case _:
	    return x;
	}
    }
    
    let t = type_map(strip_current_module, tt) in
    return Xml::Element("type", [xmlize_ty_int(t)], Nil);
}

Xml::node xmlize_ty_symbol(ty_symbol sym)
{
    return Xml::Element("tydef",	
			[xmlize_tyvars(sym.a.tysym_tyvars), 
			 xmlize_ty(sym.stype)],
			[("name", symbol_name(sym))]);
    
}

<id>list unxmlize_tyvars(Xml::node n)
{
    id get_tyvar(Xml::node k)
    {
	return Xml::prop_value("name", k);
    }
    
    switch n {
    case Xml::Element("tyvars", children, _):
	return List::map(get_tyvar, children);
    case _:
	raise Invalid_argument("Xmldump::unxmlize_tyvars");
    }
}

<<'a>symbol>field unxmlize_field(Xml::node n)
{
    return ({fld_symbol = rev_qualified_name(Xml::prop_value("name", n)),
	 fld_type = unxmlize_ty(Xml::find_element("type", n)),
	 // someday-TODO: location in field
	 fld_loc = ({file = "", first_line = 0, last_line = 0,
		     first_column = 0, last_column = 0})
	});
}

ty unxmlize_ty(Xml::node n)
{
    def children_of(n)
    {
	switch n {
	case Xml::Element(_, c, _):
	    return c;
	case _:
	    raise Invalid_argument("Xmldump::unxmlize_ty");
	}
    }

    def only_child_of(n)
    {
	switch n {
	case Xml::Element(_, [x], _):
	    return x;
	case _:
	    raise Invalid_argument("Xmldump::unxmlize_ty");
	}
    }

    def unxmlize_parm(n)
    {
	switch n {
	case Xml::Element("parm", [tt], [("name", name)]):
	    return (unxmlize_ty(tt), Some(name));
	    
	case Xml::Element("parm", [tt], []):
	    return (unxmlize_ty(tt), None);
	    
	case _:
	    raise Invalid_argument("Xmldump::unxmlize_ty");
	}
    }

    switch n {
    case Xml::Element("type", [x], _):
	n = x;
    case _:
	skip;
    }
    
    switch n {
    case Xml::Element("int", _, _):
	return Int;
	
    case Xml::Element("char", _, _):
	return Char;
	
    case Xml::Element("void", _, _):
	return Void;
	
    case Xml::Element("bool", _, _):
	return Bool;
	
    case Xml::Element("float", _, _):
	return Float;
	
    case Xml::Element("string", _, _):
	return String;
	
    case Xml::Element("abstract-type", _, _):
	return Abstract_type;
	
    case Xml::Element("exn", _, _):
	return Exn;
	
    case Xml::Element("tuple", lst, _):
	return Tuple(List::map(unxmlize_ty, lst));
	
    case Xml::Element("func", _, _):
	let ret = only_child_of(Xml::find_element("ret-type", n)) in
	let parms = children_of(Xml::find_element("parms", n)) in
	let tyvars = Xml::find_element("tyvars", n) in
	return Func({ret_type = unxmlize_ty(ret),
		     parms = List::map(unxmlize_parm, parms),
		     tyvars = unxmlize_tyvars(tyvars)});
	
    case Xml::Element("tyvar", _, _):
	return Tyvar(Xml::prop_value("name", n));
    
    case Xml::Element("struct", lst, _):
	return Struct(List::map(unxmlize_field, lst));
	
    case Xml::Element("opt-struct", lst, _):
	return Opt_struct(List::map(unxmlize_field, lst));
	
    case Xml::Element("union", lst, _):
	return Union(List::map(unxmlize_field, lst));
	
    case Xml::Element("named-type", _, _):
	let parms = children_of(Xml::find_element("parms", n)) in
	let name = Xml::prop_value("name", n) in
	return Named_type({parms = List::map(unxmlize_ty, parms),
		           tydef = rev_qualified_name(name)});
    
    case Xml::Element(x, _, _):
	raise Invalid_argument("Xmldump::unxmlize_ty " + x);

    case _:
	raise Invalid_argument("Xmldump::unxmlize_ty");
    }
}

// vim:sts=4:sw=4
