/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1012 $ $LastChangedDate: 2002-09-21 20:16:14 +0200 (Sat, 21 Sep 2002) $
 *
 * Xtcheck - check whether there are none weak type variables.
 */

open Ast;
open Util;
open Tyutil;

void subst_dummies(ty t)
{
    void f(ty tt)
    {
	switch tt {
	case X_tyvar(x):
	    switch x.t {
	    case None:
		x.t = Some(Tyvar("$dummy"));
	    case Some(_):
		skip;
	    }
	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

void xtcheck_raw_expr(nl_expr ex)
{
    switch ex.raw {
    case Assign(e1, e2):
    case Simple_binop(_, e1, e2):
	xtcheck_expr(e1);
	xtcheck_expr(e2);
	
    case Simple_unop(_, e):
	xtcheck_expr(e);
    
    case Ref(_):
	skip;

    case Call(e, el):
	xtcheck_expr(e);
	List::iter(xtcheck_expr, el);

    case Tuple_expr(el):
	List::iter(xtcheck_expr, el);

    case Union_expr(u):
	xtcheck_expr(u.nx_expr);
    
    case Lambda(l):
	xtcheck_lambda(l);

    case Struct_expr(ss, _):
	List::iter(fun void (sini_expr s) (xtcheck_expr(s.sx_expr)), ss);

    case Struct_copy(e, ss, _):
	xtcheck_expr(e);
	List::iter(fun void (sini_expr s) (xtcheck_expr(s.sx_expr)), ss);

    case Dot_expr(d, _):
	xtcheck_expr(d.sx_expr);
   
    case Assign_target:
    case Char_const(_):
    case Bool_const(_):
    case Int_const(_):
    case Float_const(_):
    case String_const(_):
    case Void_const:
    case Null_const:
	skip;
    }

    switch ex.expr_type {
    case Some(t):
	subst_dummies(t);
	
    case None:
	ice("Xtcheck::xtcheck_raw_expr(): no type?");
    }
}

void xtcheck_expr(expr e)
{
    locate(xtcheck_raw_expr, e);
}

void xtcheck_nl_pattern(nl_pattern p)
{
    switch p {
    case Void_pattern:
	skip;
	
    case Var_pattern(v):
	if (check_weak(v.var_type))
	    error("in pattern variable " + simple_name(v.var_symbol));
    
    case As_pattern(pat, v):
	if (check_weak(v.var_type))
	    error("in pattern variable " + simple_name(v.var_symbol));
	xtcheck_pattern(pat);
    
    case Expr_pattern(e):
	xtcheck_expr(e);

    case Tuple_pattern(lst):
	List::iter(xtcheck_pattern, lst);
    
    case Struct_pattern(sp):
	List::iter(fun (sp) (xtcheck_pattern(sp.sp_pattern)), sp.sp_entries);
    
    case Union_pattern(np):
	xtcheck_pattern(np.np_pattern);
    }
}

void xtcheck_pattern(pattern p)
{
    locate(xtcheck_nl_pattern, p);
}

void xtcheck_nl_stmt(nl_stmt s)
{
    void xtcheck_case_clause(case_clause cc)
    {
	List::iter(xtcheck_pattern, List::map(fst, cc.cases));
	List::iter(xtcheck_expr, List::map(snd, cc.cases));
	xtcheck_stmt(cc.body);
    }
    
    switch s {
    case Skip:
	skip;
    
    case Expr(e):
	xtcheck_expr(e);
    
    case Return(e):
	xtcheck_expr(e);

    case Vardef(v):
	if (check_weak(v.var_type))
	    error("in variable definition of " + simple_name(v.var_symbol));
	else
	    switch type_instance(v.var_type) {
	    case Void:
		// it is only allowed for parameters
		error("variable `" + simple_name(v.var_symbol) +
		      "' is bound to have `void' type");
	    case _:
		skip;
	    }
	
	
    case Block(lst):
	List::iter(xtcheck_stmt, lst);

    case Pair(s1, s2):
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Switch(e, clauses):
	xtcheck_expr(e);
	List::iter(xtcheck_case_clause, clauses);
	
    case Flying_let(assigns):
    case Let(assigns, _):
	void check_one(*(pattern, expr) a)
	{
	    let (p, e) = a in {
		xtcheck_expr(e);
		xtcheck_pattern(p);
	    }
	}

	List::iter(check_one, assigns);

	switch s {
	case Let(_, st):
	    xtcheck_stmt(st);
	case _:
	    skip;
	}
    
    case If(e, s1, s2):
	xtcheck_expr(e);
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Loop(l):
	xtcheck_stmt(l.ll_body);
	xtcheck_stmt(l.ll_inc);
    
    case Break(_):
	skip;
	
    case Break_if_not(e):
	xtcheck_expr(e);
    
    case Open(_):
	skip;
    
    case Try(tr):
	xtcheck_stmt(tr.try_body);
	List::iter(xtcheck_case_clause, tr.try_with);
	xtcheck_stmt(tr.try_finally);
    
    case Raise(e):
	xtcheck_expr(e);
    }
}

void xtcheck_stmt(stmt s)
{
    locate(xtcheck_nl_stmt, s);
}

void xtcheck_lambda(lambda_record l)
{
    List::iter(fun void (var_symbol s) {
		    if (check_weak(s.stype))
			error("in lambda parameter " + s.name);
		}, 
	       l.lambda_parms);
    if (check_weak(l.lambda_type))
	error("in lambda expression type");
    xtcheck_stmt(l.lambda_body);
}

void xtcheck_nl_decl(nl_decl nld)
{
    switch nld {
    case Func_decl(fn):
	xtcheck_lambda(fn.fn_lambda);
    
    case Section_decl(_, l):
	xtcheck_lambda(l);
    
    case Global_decl(v):
    case Extern_c(_, v):
    case Func_proto(v):
    case Exception_decl(v):
	if (check_weak(v.var_type))
	    error("in global symbol " + qualified_name(v.var_symbol));
	
    case Tydef(ts):
	if (check_weak(ts.stype))
	    error("in type definition");
	
    case Top_level_open(_):
	skip;
    }
}

void xtcheck_decl(decl d)
{
    locate(xtcheck_nl_decl, d);
}

// vim:sts=4:sw=4
