/**
+ RPN calculator with variables

We would like to extend our RPN calculator to support variables.
Variables will be set by "s" command.  It pops number and variable
name off the stack and sets this variable to given number.  Additionally
variable names can be entered where normally numbers could be.

In order to support variables we need to store both numbers and
strings on our stack.  As said before it is impossible to store
integer on string stack or vice versa.  Therfore we will need to
introduce new data type: an union.

Just like in C union can hold one of values listed in its definition.  
However unlike in C, Gont runtime system can always tell which exactly
value is stored in union at the very moment.

Union definition is similar to the same construct in C.  One needs to
list type-name pairs for each field, each terminated by ';'.  There
is convention to start names of field in union with uppercase letter.
The same convention applies to exceptions, as they are conceptually
fields of one big union.
*/

union slot {			// {0}
	int Number;
	string Variable;
}

<slot>Stack::t stack = Stack::create();			// {1}
<int>Hashtbl::t variable_bindings = Hashtbl::create(0);	// {2}

// Raised in case user causes an error.  Argument is human-readable
// error message.
exception string Error;

// Push number on the stack.
void push(int x)
{
	Stack::push(stack, Number(x));		// {3}
}

// Pop slot off the stack and convert it to number.  Raise [Error]
// if it's impossible.
int pop()
{
	try {
		slot s = Stack::pop(stack);
		
		switch s {		// {4}
		case Number(x):
			return x;
			
		case Variable(n):
			try {
				// {5}
				return Hashtbl::find(variable_bindings, n);
			} with {
			case Not_found:
				raise Error("no such variable: " + n);
			}
		}
	} with {
	case Invalid_argument(_):
		raise Error("empty stack");
	}
}

// Given a command entered by the user perform operations on stack.
void perform_operation(string cmd)
{
	int v1, v2;
	
	switch cmd {
	case "+":	push(pop() + pop());
	case "*":	push(pop() * pop());
	case "-":
		v1 = pop();
		v2 = pop();
		push(v2 - v1);
	case "/":
		v1 = pop();
		v2 = pop();
		push(v2 / v1);
	case "p":
		v1 = pop();
		puts(itoa(v1) + "\n");
		push(v1);
	case "q":
		// Here we fake end of file when user says good bye.
		raise End_of_file;
	case "s":
		v1 = pop();
		try {
			switch Stack::pop(stack) {
			case Variable(n):
				// {6}
				Hashtbl::add(variable_bindings, n, v1);

			case Number(_):		// {7}
				raise Error("expected variable, found number");
			}
		} with {
		case Invalid_argument(_):
			raise Error("empty stack");
		}
		
	// If it's not any of the above ...
	case _:
		try {
			// ... it can be number.
			push(atoi(cmd));
		} with {
		case Invalid_argument(_):
			// But if it's not, it's variable name.
			Stack::push(stack, Variable(cmd));
		}
	}
}

void main()
{
	for (;;)
		try {
			perform_operation(read_string());
		} with {
		case Error(msg):
			puts("Error: " + msg + ".\n");
		case End_of_file:
			puts("Bye.\n");
			Sys::exit(0);
		}
}

section init
{
	main();
}

/**
{0} Here we define union [slot].  Slot is either a bare integer (Number)
or variable name (Variable).

{1} Name of union, after definition, is valid type name.  There is no need
to prefix it with "union", or do something like [typedef union foo foo;]
like in C.  So we use our [slot] union in [stack] definition.

{2} [<'a>Hashtbl::t] is an associative array indexed with strings.
I.e. one can add key/value pairs, where key is string and value of
,,something'' and then lookup values based on keys.  Here we need
string -> int mapping for variables so we create one.  [0] argument
to [Hashtbl::create] means to use default size of hash table, it
will grow as neccessary.

{3} This is how union expressions are created.  This looks like a ,,call''
to union field: field name, '(', list of union field arguments (as declared
in union defintion), and ')'.  

{4} Union fields are accessed through [switch] statement.  There is no
[s.Number] notation for union fields.  This allows for easy handling
of all possible cases.  Remember that failing to match value at runtime
causes [Match_failure] exception.

If you fail to provide matches for some fields, compiler will warn you
about it.  This warning comes in very handy when you add new field to
your union, and few places in program needs to be modified.  In such
case the warning will help you pinpoint these places.

Things right after [case] statements are called patterns.  A pattern
can be integer, string, [true], [false], variable, match all pattern
(spelled [_]), union field followed by a pattern, tuple pattern (which
is list of comma-separated patterns surrounded by [()]) or struct pattern.

We already seen string patterns and match-all patterns in action in 
previous examples.  You probably can figure out how integer
and [true]/[false] patterns work :-)

Here we introduce variable patterns and union patterns.
Variable pattern matches any value and binds it to this variable, so
it can be later on referenced from [case] body.  Union pattern X matches
union value if it contains that very field.

Patterns are checked in top-to-bottom order.  In this example it means
that [s] is first checked wheteher it contains [Number] field.  If it does, [x]
is bound to value of that field, which is integer, case body is executed
and control leaves [switch].
Otherwise [s] is checked against [Variable], and in case of match
[n] is bound to string parameter of it.  Note that there is no value
that fails to match one of these cases, therefore there is no warning
from compiler.  Variables in [case] patterns are only bound within
this [case] body.  In our example it is not possible to reference
[x] after [case Variable(n):].

{5} [Hashtbl::find(h, k)] tries to lookup key (string) [k] in hashtable
[h].  If it fails [Not_found] exception is raised, otherwise value 
associated with [k] is returned.  We translate [Not_found] exception
to human readable error message.

{6} [Hashtbl::add(h, k, v)] inserts value [v] with key [k] into hash table
[h].

{7} When we need top pop variable name off top of the stack, Gont compiler
insists on checking whether it is [Variable] or [Number].  In case it is
Number, we're not really interested in what exactly number it is,
so we use [_] instead of variable name as [Number] argument.  [_] in patterns
is generally used to match anything and throw away whatever it was. 

We could have used some variable name, like [x] as well, but we would
get "unused pattern variable" warning then.
*/
